#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <pthread.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include "IOTCClient.h"
#include "AVClient.h"
#include "P2PCam/AVFRAMEINFO.h"
#include "P2PCam/AVIOCTRLDEFs.h"
#include "sample_constants.h"


#define ENABLE_RESEND 1
#define ENABLE_DTLS 0
#define ENABLE_RECV_LIVE 0 //0=only send vsaas info to device, 1=recv live video
#define VIDEO_BUF_SIZE 128000
#define VSAAS_STAGE_DOMAIN "asia-vpapi-tutk-stg.kalay.us"
#define VSAAS_PROC_DOMAIN "asia-vsaasapi-tutk.kalayservice.com"

char gAvAccount[NEW_MAXSIZE_VIEWACC + 1] = "admin";
char gAvPassword[NEW_MAXSIZE_VIEWPWD + 1] = "888888";
char gIotcAuthkey[IOTC_AUTH_KEY_LENGTH + 1] = "00000000"; //use "00000000" as default authkey
char gUID[20 + 1] = {0};
char gStoken[1024] = {0};

struct timeval gConnectStartTv;

typedef struct _AV_Server
{
	char UID[21];
	int SID;
	int twoWayStream;
	int avIndex;
} AV_Server;

//########################################################
//#     Print error message
//########################################################
void PrintErrHandling(int nErr)
{
	switch (nErr)
	{
	case IOTC_ER_SERVER_NOT_RESPONSE:
		//-1 IOTC_ER_SERVER_NOT_RESPONSE
		printf("[Error code : %d]\n", IOTC_ER_SERVER_NOT_RESPONSE);
		printf("Master doesn't respond.\n");
		printf("Please check the network wheather it could connect to the Internet.\n");
		break;
	case IOTC_ER_FAIL_RESOLVE_HOSTNAME:
		//-2 IOTC_ER_FAIL_RESOLVE_HOSTNAME
		printf("[Error code : %d]\n", IOTC_ER_FAIL_RESOLVE_HOSTNAME);
		printf("Can't resolve hostname.\n");
		break;
	case IOTC_ER_ALREADY_INITIALIZED:
		//-3 IOTC_ER_ALREADY_INITIALIZED
		printf("[Error code : %d]\n", IOTC_ER_ALREADY_INITIALIZED);
		printf("Already initialized.\n");
		break;
	case IOTC_ER_FAIL_CREATE_MUTEX:
		//-4 IOTC_ER_FAIL_CREATE_MUTEX
		printf("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_MUTEX);
		printf("Can't create mutex.\n");
		break;
	case IOTC_ER_FAIL_CREATE_THREAD:
		//-5 IOTC_ER_FAIL_CREATE_THREAD
		printf("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_THREAD);
		printf("Can't create thread.\n");
		break;
	case IOTC_ER_FAIL_CREATE_SOCKET:
		//-6 IOTC_ER_FAIL_CREATE_SOCKET
		printf("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_SOCKET);
		printf("Fails to create sockets.\n");
		break;
	case IOTC_ER_FAIL_SOCKET_OPT:
		//-7 IOTC_ER_FAIL_SOCKET_OPT
		printf("[Error code : %d]\n", IOTC_ER_FAIL_SOCKET_OPT);
		printf("Fails to set up socket options.\n");
		break;
	case IOTC_ER_FAIL_SOCKET_BIND:
		//-8 IOTC_ER_FAIL_SOCKET_BIND
		printf("[Error code : %d]\n", IOTC_ER_FAIL_SOCKET_BIND);
		printf("Fails to bind sockets.\n");
		break;
	case IOTC_ER_UNLICENSE:
		//-10 IOTC_ER_UNLICENSE
		printf("[Error code : %d]\n", IOTC_ER_UNLICENSE);
		printf("This UID is unlicense.\n");
		printf("Check your UID.\n");
		break;
	case IOTC_ER_NOT_INITIALIZED:
		//-12 IOTC_ER_NOT_INITIALIZED
		printf("[Error code : %d]\n", IOTC_ER_NOT_INITIALIZED);
		printf("Please initialize the IOTCAPI first.\n");
		break;
	case IOTC_ER_TIMEOUT:
		//-13 IOTC_ER_TIMEOUT
		break;
	case IOTC_ER_INVALID_SID:
		//-14 IOTC_ER_INVALID_SID
		printf("[Error code : %d]\n", IOTC_ER_INVALID_SID);
		printf("This SID is invalid.\n");
		printf("Please check it again.\n");
		break;
	case IOTC_ER_FAIL_GET_LOCAL_IP:
		//-16 IOTC_ER_FAIL_GET_LOCAL_IP
		printf("[Error code : %d]\n", IOTC_ER_FAIL_GET_LOCAL_IP);
		printf("Fails to get the client's local IP address.\n");
		break;
	case IOTC_ER_EXCEED_MAX_SESSION:
		//-18 IOTC_ER_EXCEED_MAX_SESSION
		printf("[Error code : %d]\n", IOTC_ER_EXCEED_MAX_SESSION);
		printf("[Warning]\n");
		printf("The amount of session reach to the maximum.\n");
		printf("It cannot be connected unless the session is released.\n");
		break;
	case IOTC_ER_CAN_NOT_FIND_DEVICE:
		//-19 IOTC_ER_CAN_NOT_FIND_DEVICE
		printf("[Error code : %d]\n", IOTC_ER_CAN_NOT_FIND_DEVICE);
		printf("Device didn't register on server, so we can't find device.\n");
		printf("Please check the device again.\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_SESSION_CLOSE_BY_REMOTE:
		//-22 IOTC_ER_SESSION_CLOSE_BY_REMOTE
		printf("[Error code : %d]\n", IOTC_ER_SESSION_CLOSE_BY_REMOTE);
		printf("Session is closed by remote so we can't access.\n");
		printf("Please close it or establish session again.\n");
		break;
	case IOTC_ER_REMOTE_TIMEOUT_DISCONNECT:
		//-23 IOTC_ER_REMOTE_TIMEOUT_DISCONNECT
		printf("[Error code : %d]\n", IOTC_ER_REMOTE_TIMEOUT_DISCONNECT);
		printf("We can't receive an acknowledgement character within a TIMEOUT.\n");
		printf("It might that the session is disconnected by remote.\n");
		printf("Please check the network wheather it is busy or not.\n");
		printf("And check the device and user equipment work well.\n");
		break;
	case IOTC_ER_DEVICE_NOT_LISTENING:
		//-24 IOTC_ER_DEVICE_NOT_LISTENING
		printf("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_LISTENING);
		printf("Device doesn't listen or the sessions of device reach to maximum.\n");
		printf("Please release the session and check the device wheather it listen or not.\n");
		break;
	case IOTC_ER_CH_NOT_ON:
		//-26 IOTC_ER_CH_NOT_ON
		printf("[Error code : %d]\n", IOTC_ER_CH_NOT_ON);
		printf("Channel isn't on.\n");
		printf("Please open it by IOTC_Session_Channel_ON() or IOTC_Session_Get_Free_Channel()\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_FAIL_CONNECT_SEARCH:
		//-27 IOTC_ER_FAIL_CONNECT_SEARCH
		printf("[Error code : %d]\n", IOTC_ER_FAIL_CONNECT_SEARCH);
		printf("The client stop connecting to the device.\n");
		break;
	case IOTC_ER_SESSION_NO_FREE_CHANNEL:
		//-31 IOTC_ER_SESSION_NO_FREE_CHANNEL
		printf("[Error code : %d]\n", IOTC_ER_SESSION_NO_FREE_CHANNEL);
		printf("All channels are occupied.\n");
		printf("Please release some channel.\n");
		break;
	case IOTC_ER_TCP_TRAVEL_FAILED:
		//-32 IOTC_ER_TCP_TRAVEL_FAILED
		printf("[Error code : %d]\n", IOTC_ER_TCP_TRAVEL_FAILED);
		printf("Device can't connect to Master.\n");
		printf("Don't let device use proxy.\n");
		printf("Close firewall of device.\n");
		printf("Or open device's TCP port 80, 443, 8080, 8000, 21047.\n");
		break;
	case IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED:
		//-33 IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED
		printf("[Error code : %d]\n", IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED);
		printf("Device can't connect to server by TCP.\n");
		printf("Don't let server use proxy.\n");
		printf("Close firewall of server.\n");
		printf("Or open server's TCP port 80, 443, 8080, 8000, 21047.\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_NO_PERMISSION:
		//-40 IOTC_ER_NO_PERMISSION
		printf("[Error code : %d]\n", IOTC_ER_NO_PERMISSION);
		printf("This UID's license doesn't support TCP.\n");
		break;
	case IOTC_ER_NETWORK_UNREACHABLE:
		//-41 IOTC_ER_NETWORK_UNREACHABLE
		printf("[Error code : %d]\n", IOTC_ER_NETWORK_UNREACHABLE);
		printf("Network is unreachable.\n");
		printf("Please check your network.\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_FAIL_SETUP_RELAY:
		//-42 IOTC_ER_FAIL_SETUP_RELAY
		printf("[Error code : %d]\n", IOTC_ER_FAIL_SETUP_RELAY);
		printf("Client can't connect to a device via Lan, P2P, and Relay mode\n");
		break;
	case IOTC_ER_NOT_SUPPORT_RELAY:
		//-43 IOTC_ER_NOT_SUPPORT_RELAY
		printf("[Error code : %d]\n", IOTC_ER_NOT_SUPPORT_RELAY);
		printf("Server doesn't support UDP relay mode.\n");
		printf("So client can't use UDP relay to connect to a device.\n");
		break;
	case IOTC_ER_NO_SERVER_LIST:
		//-44 IOTC_ER_NO_SERVER_LIST
		printf("[Error code : %d]\n", IOTC_ER_NO_SERVER_LIST);
		printf("No IOTC server information while client connect.\n");
		break;
	case IOTC_ER_DEVICE_MULTI_LOGIN:
		//-45 IOTC_ER_DEVICE_MULTI_LOGIN
		printf("[Error code : %d]\n", IOTC_ER_DEVICE_MULTI_LOGIN);
		printf("The connecting device duplicated loggin and may unconnectable.\n");
		break;
	case IOTC_ER_MASTER_NOT_RESPONSE:
		//-60 IOTC_ER_MASTER_NOT_RESPONSE
		printf("[Error code : %d]\n", IOTC_ER_MASTER_NOT_RESPONSE);
		printf("All masters have no respond.\n");
		break;
	case IOTC_ER_DEVICE_IS_SLEEP:
		//-64 IOTC_ER_DEVICE_IS_SLEEP
		printf("[Error code : %d]\n", IOTC_ER_DEVICE_IS_SLEEP);
		printf("Device is in sleep mode.\n");
		break;
	case IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY:
		// -68 IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY
		printf("[Error code : %d]\n", IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY);
		printf("Being rejected by device due to the wrong auth key.\n");
		break;
	case IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION:
		// -69 IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION
		printf("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION);
		printf("Being rejected by device due to disable authentication.\n");
		break;

	default:
		printf("[Unknow error code : %d]\n", nErr);
		break;
	}
}

//########################################################
//#     Get VSaaS information from app
//########################################################
void GetVsaasInfoFromApp(char *vsaas_info)
{
    //not implement in this sample
	// In this sample code, we assign the VSAAS_STAGE_DOMAIN as a domain name,
	// If you wanna change the domain name to use kalay app, you can use VSAAS_PROC_DOMAIN to replace it.
    sprintf(vsaas_info, "{\"func\":\"setCloudRecordingEndpoint\",\"args\":{\"url\":\"https://%s/vsaas/api/v1/stream/stream_url/%s?stoken=%s\"}}", VSAAS_STAGE_DOMAIN, gUID, gStoken);
}

//########################################################
//#     Send IOCtrl CMD to device
//########################################################

void ConnectDevcie(AV_Server *ServerInfo)
{
	int ret, SID, tmpSID, avIndex;
	char vsaas_info[1024] = {0};

	tmpSID = IOTC_Get_SessionID();
	if (tmpSID < 0)
	{
		printf("  [%s:%d] IOTC_Get_SessionID failed ret[%d]\n", __FUNCTION__, __LINE__, tmpSID);
		pthread_exit(0);
	}
	printf("  [%s:%d] IOTC_Get_SessionID tmpSID[%d]\n", __FUNCTION__, __LINE__, tmpSID);

	IOTCConnectInput connect_input = {0};
	connect_input.cb = sizeof(connect_input);
	connect_input.authentication_type = AUTHENTICATE_BY_KEY;
	connect_input.timeout = 20; //sec
	memcpy(connect_input.auth_key, gIotcAuthkey, IOTC_AUTH_KEY_LENGTH);

	SID = IOTC_Connect_ByUIDEx(ServerInfo->UID, tmpSID, &connect_input);

	if (SID < 0)
	{
		printf("  [%s:%d] IOTC_Connect_ByUID_Parallel failed SID[%d]\n", __FUNCTION__, __LINE__, SID);
		PrintErrHandling(SID);
		pthread_exit(0);
	}
	else
	{
		struct timeval tv2;
		gettimeofday(&tv2, NULL);
		unsigned long msec = (tv2.tv_sec - gConnectStartTv.tv_sec) * 1000 + (tv2.tv_usec - gConnectStartTv.tv_usec) / 1000;
		printf("  [%s:%d] IOTC_Connect_ByUID_Parallel to %s SID[%d] Cost Time[%lu]ms\n", __FUNCTION__, __LINE__, ServerInfo->UID, SID, msec);

		struct st_SInfoEx Sinfo;
        memset(&Sinfo, 0, sizeof(struct st_SInfoEx));
        Sinfo.size = sizeof(struct st_SInfoEx);
        if (IOTC_Session_Check_Ex(SID, &Sinfo) == IOTC_ER_NoERROR)
		{
			char *mode[] = {"P2P", "RLY", "LAN"};
			if (isdigit(Sinfo.RemoteIP[0]))
				printf("  [%s:%d] Device[%s] is from [%s:%d] Mode[%s] NatType[%d] IOTCVersion[%X]\n", __FUNCTION__, __LINE__,
					   Sinfo.UID, Sinfo.RemoteIP, Sinfo.RemotePort, mode[(int)Sinfo.Mode], Sinfo.LocalNatType, Sinfo.IOTCVersion);
		}
	}

	AVClientStartInConfig avClientInConfig;
	AVClientStartOutConfig avClientOutConfig;

	avClientInConfig.cb = sizeof(AVClientStartInConfig);
	avClientInConfig.iotc_channel_id = 0;
	avClientInConfig.iotc_session_id = SID;
	avClientInConfig.timeout_sec = 30;
	avClientInConfig.resend = ENABLE_RESEND;
	avClientInConfig.security_mode = ENABLE_DTLS ? AV_SECURITY_DTLS : AV_SECURITY_SIMPLE;
	avClientInConfig.account_or_identity = gAvAccount;
	avClientInConfig.password_or_token = gAvPassword;
	avClientInConfig.auth_type = AV_AUTH_PASSWORD;
	avClientOutConfig.cb = sizeof(AVClientStartOutConfig);

	avIndex = avClientStartEx(&avClientInConfig, &avClientOutConfig);
	if (avIndex < 0)
	{
		printf("  [%s:%d] avClientStartEx failed avIndex[%d]\n", __FUNCTION__, __LINE__, avIndex);
		pthread_exit(0);
	}
	printf("  [%s:%d] avClientStartEx success avIndex[%d] resend[%d] two_way_streaming[%d]\n", __FUNCTION__, __LINE__,
		   avIndex, avClientOutConfig.resend, avClientOutConfig.two_way_streaming);

	//get vsaas info & send to device

    NebulaJsonObject *response_obj = NULL;
    GetVsaasInfoFromApp(vsaas_info);
    //ret = avSendJSONCtrl(avIndex, vsaas_info, 10);
    ret = avSendJSONCtrlRequest(avIndex, vsaas_info, &response_obj, 10);
    avFreeJSONCtrlResponse(response_obj);

    printf("  [%s:%d] ret[%d]\n", __FUNCTION__, __LINE__, ret);
	avClientStop(avIndex);
	printf("  [%s:%d] avClientStop avIndex[%d]\n", __FUNCTION__, __LINE__, avIndex);
	IOTC_Session_Close(SID);
	printf("  [%s:%d] IOTC_Session_Close SID[%d]\n", __FUNCTION__, __LINE__, SID);
	return;
}

//########################################################
//#  Main function of this sample client
//########################################################
int main(int argc, char *argv[])
{
	int ret = 0;

    if (argc < 3) {
        printf("Usage: %s UID stoken\n", argv[0]);
		return 0;
	}
    strncpy(gUID, argv[1], 20);
    strncpy(gStoken, argv[2], 1023);

    printf("%s start\n", argv[0]);
    printf("IOTCAPI version[%s] AVAPI version[%s]\n", IOTC_Get_Version_String(), avGetAVApiVersionString());
    srand(time(NULL));

    const char *license_key = getenv(ENV_LICENSE_KEY);
    if (license_key == NULL) {
        printf("Environment variable %s must be set!!\n", ENV_LICENSE_KEY);
        return -1;
    }
    ret = TUTK_SDK_Set_License_Key(license_key);
    if (ret != TUTK_ER_NoERROR)
    {
        printf("TUTK_SDK_Set_License_Key()=%d ,exit...!!\n", ret);
        return 0;
    }

	gettimeofday(&gConnectStartTv, NULL);
	ret = IOTC_Initialize2(0);
	if (ret != IOTC_ER_NoERROR)
	{
		printf("[%s:%d] IOTC_Initialize2 failed ret[%d]\n", __FUNCTION__, __LINE__, ret);
		PrintErrHandling(ret);
		return -1;
	}

	// alloc 3 sessions for video and two-way audio
	avInitialize(32);

	AV_Server ServerInfo;

	memset(&ServerInfo, 0, sizeof(AV_Server));
    strncpy(ServerInfo.UID, gUID, 21);

	ConnectDevcie(&ServerInfo);

	avDeInitialize();
	IOTC_DeInitialize();

	printf("%s exit\n", argv[0]);

	return 0;
}
